# Python
import json
from typing import Union, Callable

# Django
from django.http.response import HttpResponse

# Third Party
from typing import Optional, Any

# Project
from .choices import ResponseChoices


class NotLoggedResponse(HttpResponse):

    def __init__(self):
        super().__init__(json.dumps({'result': 'not logged'}), status=401)


class APIResponse(HttpResponse):

    def __init__(self, response, request, status=200):
        if request.token is not None and request.token != '':
            super().__init__(json.dumps(response, default=str), content_type='application/json', status=status)
        else:
            super().__init__(json.dumps(response, default=str), status=status)


class NotAllowedResponse(HttpResponse):

    def __init__(self, reason: Optional[str] = None):
        super().__init__(json.dumps({'result': 'not allowed', 'reason': reason or ""}), status=405)


class VieoloResponse:

    def __init__(
        self, 
        result: str, 
        object_key: str = "",
        obj: str | None = None, 
        type_of_object: str | None = None, 
        operation: str | None = None, 
        reason: str | None = None, 
        message: str | None = None, 
        status_code: int | None = None,
    ):
        self.result = result
        self.obj = obj
        self.type_of_object = type_of_object
        self.operation = operation
        self.reason = reason
        self.message = message
        self.response_object = {}
        self.status_code = status_code
        self.object_key = object_key

    def __str__(self) -> str:
        return str(self.response_object)

    def is_successful(self) -> bool:
        return self.result == ResponseChoices.success
    
    def is_does_not_exist(self) -> bool:
        return self.result == ResponseChoices.does_not_exist

    def is_already_exists(self) -> bool:
        return self.result == ResponseChoices.already_exists

    def is_invalid(self) -> bool:
        return self.result == ResponseChoices.not_valid

    def is_not_allowed(self) -> bool:
        return self.result == ResponseChoices.not_allowed
    
    def get_id(self) -> int:
        return self.response_object[self.object_key]["id"]

    def filter(self, func: Callable[[dict], bool]) -> list[dict]:
        return list(filter(func, self.data))

    @property 
    def data(self):
        return self.response_object[self.object_key]

    @property 
    def data_len(self) -> int:
        return len(self.response_object[self.object_key])

    @staticmethod 
    def parse(raw: Union[str, HttpResponse], object_key = "") -> 'VieoloResponse':
        if isinstance(raw, HttpResponse):
            parsed = json.loads(raw.content)
            status_code = raw.status_code
        else:
            parsed = json.loads(raw)
            status_code = None

        response = VieoloResponse(
            result=parsed['result'],
            status_code=status_code,
            object_key=object_key,
        )

        for k, v in parsed.items():
            if k == 'operation':
                response.operation = v
            elif k == 'object':
                response.obj = v
            elif k == 'type':
                response.type_of_object = v
            elif k == 'reason':
                response.reason = v
            elif k == 'message':
                response.message = v

            response.response_object[k] = v

        return response


def generate_response(result, obj=None, type_of_object=None, operation=None, reason=None, message=None) -> dict[str, Any]:
    response = {
        "result": result
    }
    if obj is not None:
        response["object"] = obj
    if type_of_object is not None:
        response["type"] = type_of_object
    if operation is not None:
        response["operation"] = operation
    if reason is not None:
        response["reason"] = reason
    if message is not None:
        response["message"] = message

    return response



from xml.etree import cElementTree
from xml.etree.ElementTree import ParseError

from typing import List, Union


class XMLParseError(ParseError):
    pass


class XMLElement:

    @staticmethod
    def parse_string(raw: str):
        """
        Parses a string a return an instance of XMLElement
        :param raw: The XML data in the string format
        :return: an Instance of XMLElement
        """
        try:
            return XMLElement(cElementTree.fromstring(raw))
        except ParseError:
            raise XMLParseError()  # noqa: B904
        except TypeError:
            raise XMLParseError()  # noqa: B904
        
    @staticmethod
    def try_parse_string(raw: str) -> Union['XMLElement', None]:
        """
        Parses a string a return an instance of XMLElement
        If the raw cannot be parsed, None is returned
        :param raw: The XML data in the string format
        :return: an Instance of XMLElement
        """
        try:
            return XMLElement.parse_string(raw)
        except XMLParseError:
            return None

    def __init__(self, element):
        self.element = element

    def __str__(self):
        return "<{t} attrib={a} childrenCount={c} />".format(t=self.tag, c=self.children_count, a=self.attrib)

    @property
    def text(self) -> str:
        """The text property of the native xml element"""
        return self.element.text

    @property
    def attrib(self) -> dict:
        """The attribute element of the native xml element"""
        return self.element.attrib

    @property
    def tag(self) -> str:
        """The name of the tag of the native xml element. Namespaces will be removed, if any."""
        splitted = self.element.tag.split("}")
        return splitted[len(splitted) - 1]

    @property
    def full_tag(self) -> str:
        """The the tag of the native xml element, including the namespaces, if any"""
        return self.element.tag

    @property
    def children(self):
        """A list containing the children of the element"""
        children: List[XMLElement] = []
        for child in list(self.element):
            children.append(XMLElement(child))
        return children

    @property
    def children_count(self) -> int:
        """The number of children"""
        return len(self.element)

    @property
    def has_child(self) -> bool:
        """Whether the element has any children"""
        return len(self.element) > 0

    @property
    def has_attrib(self) -> bool:
        """Checks if the element has any attributes"""
        return len(self.attrib) > 0

    def iter(self, target_tag: str):
        """Similar to the children property"""
        iterations = self.element.iter(target_tag)
        children: List[XMLElement] = []
        for i in iterations:
            children.append(XMLElement(i))
        return children

    def find_all(self, target_tag: str):
        """Finds all child elements whose tag name matches that of the target tag"""
        iterations = self.element.findall(target_tag)
        children: List[XMLElement] = []
        for i in iterations:
            children.append(XMLElement(i))
        return children

    def find(self, target_tag: str):
        """Finds the first child whose tag name matches that of the target tag"""
        found: XMLElement = XMLElement(self.element.find(target_tag))
        return found

    def to_dict(self) -> dict:
        """Converts the XML element and its children to a dict"""        

        x: dict = {
            "tag": self.full_tag            
        }

        if self.has_attrib:
            x["attributes"] = self.attrib

        if self.has_child:
            children = self.children
            x["children"] = []

            for c in children:
                x["children"].append(c.to_dict())        
        elif self.text is not None:
            parsed = XMLElement.try_parse_string(self.text)
            if parsed is not None:
                x['children'] = [parsed.to_dict()]
            else:
                x["text"] = self.text

        return x


# Django
from django.test import TestCase

from django_carpet.response import VieoloResponse


class ResponseTest(TestCase):

    def test_vieolo_response(self):
        vr = VieoloResponse('success', 'items')
        vr.response_object['items'] = [{"id": 1, "name": "one"}, {"id": 2, "name": "two"}, {"id": 3, "name": "three"}]
        
        self.assertEqual(len(vr.data), 3)

        filterd_data = vr.filter(lambda x: x["id"] < 3) 
        self.assertEqual(len(filterd_data), 2)
        self.assertEqual(filterd_data[0]["id"], 1)
        self.assertEqual(filterd_data[1]["id"], 2)


# Django
from django.conf import settings

# Foundation
from .choices import EnvChoices


def exec_mode() -> str:
    """Gets the `EXEC_MODE` variable from the settings, informing the current state of runtime

    Returns:
        A member of `EnvChoices`
    """
    return getattr(settings, "EXEC_MODE", EnvChoices.NORMAL)


def is_debug() -> bool:
    return settings.DEBUG


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="django_carpet",  # Replace with your own username
    version="0.1.13",
    author="Vieolo OÜ",
    description="Foundational functionalities for Django",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Vieolo/django-carpet.git",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.11',
)
